#!/usr/bin/perl
binmode STDOUT, ":utf8";

# Program Manual_Crypto

# A collection of manual crypto methods, just for fun!

# See the "about" files in the "resources" directory for additional information.

# NOTE: There are a lot of variables named tyui; these store temporary values that
# are maintained only for the verbose mode output.

# (c) Gary C. Kessler, 2017, 2018

$build_date = "06/08/2018";
$version = "0.8.05";

$ord0 = ord ("0");
$ordA = ord ("A");
$verbose = 0; # Initial state = OFF

@method = ("Caesar", "N mu mod26", "Atbash", "Vigenere Square", "Vigenere Square (Autokey Cipher)", "Beaufort Cipher", "Columnar Transposition",
           "Myszkowski Transpostion","ADFGVX","One-Time Pad");
@method_file = ("Caesar.txt", "Numod26.txt", "atbash.txt", "vigenere_keyword.txt","vigenere_autokey.txt", "beaufort.txt", "transposition.txt",
                "myszkowski.txt","adfgvx.txt","otp.txt");

# Wheatstone-Playfair and other varieties?
# Route transposition

# display output


do
  {
  ($option,$action,$verbose) = menu ();

  if ($action eq "About")
      {
      AboutFile ($method_file [$option-1]);
      }
    else
      {
      OPTIONS:
        {
        $option == 1 && do
          {
          Caesar ($action);
          last OPTIONS;
          };
        $option == 2 && do
          {
          Numod26 ($action);
          last OPTIONS;
          };
        $option == 3 && do
          {
          Atbash ($action);
          last OPTIONS;
          };
        $option == 4 && do
          {
          VigenereBeaufort ($action,"Vigenere");
          last OPTIONS;
          };
        $option == 5 && do
          {
          Autokey ($action);
          last OPTIONS;
          };
        $option == 6 && do
          {
          VigenereBeaufort ($action,"Beaufort");
          last OPTIONS;
          };
        $option == 7 && do
          {
          Transposition ($action,"Columnar");
          last OPTIONS;
          };
        $option == 8 && do
          {
          Transposition ($action,"Myszkowski");
          last OPTIONS;
          };
        $option == 9 && do
          {
          ADFGVX ($action);
          last OPTIONS;
          };
        $option == 10 && do
          {
          OneTimePad ($action);
          last OPTIONS;
          };
        $option eq "H" && do
          {
          AboutFile ("help.txt");
          last OPTIONS;
          };
        $option eq "X" && do
          {
          print "Goodbye!\n\n";
          exit;
          };
        };
      };

   print "\nHit <ENTER> to continue...\n";
   $tyui = <>;

  }
  until (0!=0);

# ********************************************************
# *****           SUBROUTINE BLOCKS                  *****
# ********************************************************

# ********************************************************
# menu
# ********************************************************

# Main menu and sub-menu subroutines

# Main menu sets encryption method ($i) and global $verbose
#   variable.

# Submenu then determines where we want to print the information
#  file, encrypt, or decrypt ($action)

sub menu ()
  {
  my ($action,$i,$opt);

# Show the main menu options...

  do
    {
    print "\033[2J";    # Clear the screen
    print "\033[0;0H";  # Place cursor at 0,0 (upper left-hand corner)
    print "                CRYPTO EMULATOR (Build: $build_date Version: $version)\n\n";
    print "                                   MENU\n\n";

    $num_method = @method;
    for ($i=0; $i<$num_method; $i++)
      {
      printf (" %2d. %s\n", $i+1, $method[$i]);
      }
    print "\n";
    print "  X. Exit\n\n";
    print "  H. Help\n";
    print "  V. Turn Verbose mode ";
    if ($verbose == 0)
        { print "ON (Verbose mode currently OFF)\n\n"; }
      else
        { print "OFF (Verbose mode currently ON)\n\n"; }

    print "Enter option (1-$num_method,H,V,X): ";
    chomp ($opt = <STDIN>);
    $opt = uc ($opt);
    }
    until (($opt >= 1 && $opt <= $num_method) || $opt eq "H" || $opt eq "V" || $opt eq "X");

  if ($opt eq "X")
    { return ($opt,"",0); };
  if ($opt eq "V")
    {
    $verbose = abs ($verbose - 1);
    return (0,"",$verbose);
    }
  if ($opt eq "H")
    { return ($opt,"",$verbose); };

# Now show the sub-menu and set some variables....

  do
    {
    print "\n\n    Sub-menu for $method[$opt-1]\n\n";
    print "  1. About $method[$opt-1]\n";
    print "  2. Encrypt data\n";
    print "  3. Decrypt data\n";
    print "\n";
    print " X. Exit\n\n";
    print "Enter option (1-3,X): ";
    chomp ($i = <STDIN>);
    $i = uc ($i);
    }
    until (($i >= 1 && $i <= 3) || $i eq "X");

  MENU:
    {
    $i == 1 && do
      {
      $action = "About";
      last MENU;
      };
    $i == 2 && do
      {
      $action = "Encrypt";
      last MENU;
      };
    $i == 3 && do
      {
      $action = "Decrypt";
      last MENU;
      };
    $i eq "X" && do
      {
      $action = "";
      $opt = "X";
      };
    };

  return ($opt,$action,$verbose);
  }

# ********************************************************
# AboutFile
# ********************************************************

# Print out the file named $filename

sub AboutFile ($filename)
  {
  my ($line);
  my $filename; $filename = $_[0];

  open (INFILE, "<", "resources/".$filename)
    or die "Can't open input file \"$filename\": $!";

  print "\n\n";

  while (not (eof INFILE))
    {
    $line = <INFILE>;
    print $line;
    } ;
  }

# ********************************************************
# Atbash
# ********************************************************

# Implements Atbash cipher.

sub Atbash ($action)
  {
  my ($i, $msg_in, $msg_out, $msglen, $tyui1, $tyui2, $block);
  my $action; $action = $_[0];

# Common code to do setup and read plaintext/ciphertext

  print "\n *** Atbash (Reverse) Cipher ***\n\n";
  printf " == %sion ==\n", $action;

# Read plaintext/ciphertext

  printf "\nEnter message to %s:\n", lc ($action);
  chomp ($msg_in=<>);

# Set up message arrays and variables

  ($msg_in, $msg_out, $msglen) = set_msg_values ($msg_in);

  if ($action eq "Encrypt")
    { $block = set_output_block_size (); }

  if ($verbose == 1) { print "\n"; }

  for ($i=0; $i<$msglen; $i++)
    {
    $tyui1 = ord (substr ($msg_in,$i,1)) - $ordA;
    $tyui2 = chr (25 - $tyui1 + $ordA);
    $msg_out = $msg_out . $tyui2;
    if ($verbose == 1)
      {
      if ($action eq "Encrypt")
          { printf "Plaintext = %s Ciphertext = %s\n", chr ($tyui1 + $ordA), $tyui2; }
        else
          { printf "Ciphertext = %s Plaintext = %s\n", chr ($tyui1 + $ordA), $tyui2; }
      }
    }

  if ($action eq "Encrypt")
      {
      print "\nEncrypted message = ";
      cipher_out ($msg_out,$block);
      print "\n";
      }
    else
      { print "\nDecrypted message = $msg_out\n"; }
  }

# ********************************************************
# Autokey
# ********************************************************

# Implements Autokey version of Vigenere Square.

# Program requests a priming key and then performs encrytion
#   or decryption steps.

sub Autokey ($action)
  {
  my ($i, $key, $keynum, $msg_in, $msg_out, $msglen, $tyui1, $tyui2, $block);
  my $action; $action = $_[0];

# Common code to do setup and read plaintext/ciphertext

  print "\n *** Vignere Square (Autokey Cipher) ***\n\n";
  printf " == %sion ==\n", $action;

# Get priming key

  do
    {
    print "\nEnter priming key (A-Z): ";
    chomp ($key = <STDIN>);
    $key = uc ($key);
    }
    until ($key >= "A" && $key <= "Z");

  $keynum = ord ($key) - $ordA;

# Set alphabet. In this method, the Baseword is null

  set_alphabet ("");

# Read plaintext/ciphertext

  printf "\nEnter message to %s:\n", lc ($action);
  chomp ($msg_in=<>);

# Set up message arrays and variables

  ($msg_in, $msg_out, $msglen) = set_msg_values ($msg_in);

  if ($action eq "Encrypt")
      {
      $block = set_output_block_size ();

      if ($verbose == 1)
          { print "\n"; }

      for ($i=0; $i<$msglen; $i++)
        {
        $tyui1 = ord (substr ($msg_in,$i,1)) - $ordA;
        $tyui2 = ($alphanum [$tyui1] + $alphanum [$keynum]) % 26;
        $msg_out = $msg_out . $alphabet [$tyui2];
        if ($verbose == 1)
          { printf "Plaintext = %s Key character = %s Ciphertext = %s\n", chr ($tyui1 + $ordA), chr ($keynum + $ordA), $alphabet [$tyui2]; } 
        $keynum = $tyui1;
        }

      print "\nEncrypted message = ";
      cipher_out ($msg_out,$block);
      print "\n";
      }
    else
      {
      if ($verbose == 1)
          { print "\n"; }

      for ($i=0; $i<$msglen; $i++)
        {
        $tyui1 = ord (substr ($msg_in,$i,1)) - $ordA;
        $tyui2 = ($alphanum [$tyui1] - $alphanum [$keynum]) % 26;
        $msg_out = $msg_out . $alphabet [$tyui2];
        if ($verbose == 1)
          { printf "Ciphertext = %s Key character = %s Plaintext = %s\n", chr ($tyui1 + $ordA), chr ($keynum + $ordA), $alphabet [$tyui2]; } 
        $keynum = ord ($alphabet [$tyui2]) - $ordA;
        }

      print "\nDecrypted message = $msg_out\n";
      }
  }

# ********************************************************
# Caesar
# ********************************************************

# Implements Caesar cipher.

# Program requests a shift value (N) and then performs encrytion
#   or decryption steps.

sub Caesar ($action)
  {
  my ($i, $msg_in, $msg_out, $msglen, $n, $tyui1, $tyui2, $block);
  my $action; $action = $_[0];

# Common code to do setup and read plaintext/ciphertext

  print "\n *** Caesar Cipher ***\n\n";
  printf " == %sion ==\n", $action;

# Read plaintext/ciphertext

  printf "\nEnter message to %s:\n", lc ($action);
  chomp ($msg_in=<>);

# Set up message arrays and variables

  ($msg_in, $msg_out, $msglen) = set_msg_values ($msg_in);

# Ask for N where + value is a forward shift and - value is backward

  do
    {
    print "\nEnter number of characters to shift (-25 to 25): ";
    chomp ($n = <>);
    }
    until ($n >= -25 && $n <= 25);

# Get output blocksize

  if ($action eq "Encrypt")
    { $block = set_output_block_size (); }

  if ($verbose == 1) { print "\n"; }

  for ($i=0; $i<$msglen; $i++)
    {
    $tyui1 = ord (substr ($msg_in,$i,1)) - $ordA;
    $tyui2 = chr ((($tyui1 + $n) % 26) + $ordA);
    $msg_out = $msg_out . $tyui2;
    if ($verbose == 1)
      {
      if ($action eq "Encrypt")
          { printf "Plaintext = %s Ciphertext = %s\n", chr ($tyui1 + $ordA), $tyui2; }
        else
          { printf "Ciphertext = %s Plaintext = %s\n", chr ($tyui1 + $ordA), $tyui2; }
      }
    }

  if ($action eq "Encrypt")
      {
      print "\nEncrypted message = ";
      cipher_out ($msg_out,$block);
      print "\n";
      }
    else
      { print "\nDecrypted message = $msg_out\n"; }
  }

# ********************************************************
# Numod26
# ********************************************************

# Implements Numod26 cipher.

# Program requests a multiplier value (N) and then performs encrytion
#   or decryption steps.

sub Numod26 ($action)
  {
  my ($i, $msg_in, $msg_out, $msglen, $n, $tyui1, $tyui0, $tyui2, $block);
  my $action; $action = $_[0];

# Common code to do setup and read plaintext/ciphertext

  printf "\n *** N mu mod26 Cipher ***\n\n";
  printf " == %sion ==\n", $action;

# Read plaintext/ciphertext

  printf "\nEnter message to %s:\n", lc ($action);
  chomp ($msg_in=<>);

# Set up message arrays and variables

  ($msg_in, $msg_out, $msglen) = set_msg_values ($msg_in);

# Ask for N; must be odd, between 3-25 and not 13!

  do
    {
    print "\nEnter multiplier (N) value (odd value between 3 and 25, not 13): ";
    chomp ($n = <>);
    }
    until ($n >= 3 && $n <= 25 && $n%2 != 0 && $n != 13);

# Get output blocksize

  if ($action eq "Encrypt")
    { $block = set_output_block_size (); }

  if ($verbose == 1) { print "\n"; }

  for ($i=0; $i<$msglen; $i++)
    {
  if ($action eq "Encrypt")
      {
      $tyui1 = ord (substr ($msg_in,$i,1)) - $ordA;
      $tyui2 = (($tyui1 * $n) % 26);
      }
    else
      {
      $tyui1 = ord (substr ($msg_in,$i,1)) - $ordA;

# We keep $tyui0 only for verbose mode!!!!

      $tyui0 = $tyui1;
      while ($tyui1 % $n != 0)
        { $tyui1 += 26; }
      $tyui2 = ($tyui1 / $n);
      }
    $msg_out = $msg_out . chr ($tyui2 + $ordA);
    if ($verbose == 1)
      {
      if ($action eq "Encrypt")
          { printf "Plaintext = %s (%2d) Ciphertext = %s (%2d)\n", chr ($tyui1 + $ordA), $tyui1, chr ($tyui2 + $ordA), $tyui2; }
        else
          { printf "Ciphertext = %s (%2d-->%3d) Plaintext = %s (%2d)\n", chr ($tyui0 + $ordA), $tyui0, $tyui1, chr ($tyui2 + $ordA), $tyui2; }
      }
    }

  if ($action eq "Encrypt")
      {
      print "\nEncrypted message = ";
      cipher_out ($msg_out,$block);
      print "\n";
      }
    else
      { print "\nDecrypted message = $msg_out\n"; }
  }

# ********************************************************
# VigenereBeaufort
# ********************************************************

# Implements Vigenere Square and Beaufort Cipher, two types
#   of polyalphabetic cipher using a keyword.

# Beaufort Cipher variant employs a baseword to set the
#   alphabet as well as a keyword for encryption.

# Program requests the baseword and keyword, and then performs
#   encrytion or decryption steps.

sub VigenereBeaufort ($action,$type)
  {
  my ($i, $k, $baseword, $keyword, $keylen, $keynum, $msg_in, $msg_out, $msglen, $tyui1, $tyui2, $block);
  my $action, $type; $action = $_[0]; $type = $_[1];

# Common code to do setup and read plaintext/ciphertext

  print "\n *** $type Cipher ***\n\n";
  printf " == %sion ==\n", $action;

# Get keyword and baseword (only if Beaufort)

  printf "\nEnter keyword (for %sion): ", lc ($action);
  chomp ($keyword = <STDIN>);
  $keyword = uc ($keyword);
  $keylen = length ($keyword);

  if ($type eq "Beaufort")
      {
      print "\nEnter baseword (for seeding the alphabet): ";
      chomp ($baseword = <STDIN>);
      $baseword = uc ($baseword);
      }
    else
      { $baseword = ""; }

# Set alphabet using baseword

  set_alphabet ($baseword);

# Read plaintext/ciphertext

  printf "\nEnter message to %s:\n", lc ($action);
  chomp ($msg_in=<>);

# Set up message arrays and variables

  ($msg_in, $msg_out, $msglen) = set_msg_values ($msg_in);

  if ($action eq "Encrypt")
      {
      $block = set_output_block_size ();

      $k = 0;
      if ($verbose == 1)
          { print "\n"; }

      for ($i=0; $i<$msglen; $i++)
        {
        $tyui1 = ord (substr ($msg_in,$i,1)) - $ordA;
        $keynum = ord (substr ($keyword,$k,1)) - $ordA;
        $tyui2 = ($alphanum [$tyui1] + $alphanum [$keynum]) % 26;
        $msg_out = $msg_out . $alphabet [$tyui2];
        if ($verbose == 1)
          { printf "Plaintext = %s Key character = %s Ciphertext = %s\n", chr ($tyui1 + $ordA), chr ($keynum + $ordA), $alphabet [$tyui2]; } 
        $k = ($k + 1) % $keylen;
        }

      print "\n\nEncrypted message = ";
      cipher_out ($msg_out,$block);
      print "\n";
      }
    else
      {
      $k = 0;
      if ($verbose == 1)
          { print "\n"; }

      for ($i=0; $i<$msglen; $i++)
        {
        $tyui1 = ord (substr ($msg_in,$i,1)) - $ordA;
        $keynum = ord (substr ($keyword,$k,1)) - $ordA;
        $tyui2 = ($alphanum [$tyui1] - $alphanum [$keynum]) % 26;
        $msg_out = $msg_out . $alphabet [$tyui2];
        if ($verbose == 1)
          { printf "Ciphertext = %s Key character = %s Plaintext = %s\n", chr ($tyui1 + $ordA), chr ($keynum + $ordA), $alphabet [$tyui2]; } 
        $k = ($k + 1) % $keylen;
        }

      print "\n\nDecrypted message = $msg_out\n";
      }
  }

# ********************************************************
# ADFGVX
# ********************************************************

# ADFGVX uses a 6x6 array of 36 characters. Each character
# in the plaintext message maps to the row/column intersection
# in the matrix, yielding a pair of characters.

# The new message is handled like a Myszkowski transpostion
# using a keyword.

sub ADFGVX ($action)
  {
  my ($adfgvx, $array, $block, $col, $i, $j, $k, $keyword, $mix, $msg_in, $msg_out, $msg_temp, $num_col, $row);
  my $action; $action = $_[0];

  $adfgvx = "ADFGVX";

# Fill array as a scalar although we will use as a 2D array

  $array = "FL1AO2" . "JDW3GU" . "CIYB4P" . "R5Q8VE" . "6K7ZMX" . "SNH0T9";

# This is "actual" 6x6 array:
# "F","L","1","A","O","2"
# "J","D","W","3","G","U"
# "C","I","Y","B","4","P"
# "R","5","Q","8","V","E"
# "6","K","7","Z","M","X"
# "S","N","H","0","T","9"

  print "\n ** ADFGVX Operation**\n\n";
  printf " == %sion ==\n", $action;

# Common code for setup and I/O

# Start by getting the transposition keyword and setting the column
#   transpostion sort order

  print "\nEnter transposition keyword: ";
  chomp ($keyword = <STDIN>);
  $keyword = uc ($keyword);
  $keyword =~ s/[^A-Z0-9]//g;
  $num_col = length ($keyword);
  $mix = sort_key ($keyword);

  if ($verbose == 1)
    {
    print "\n*** keyword $keyword order = $mix\n\n";

    print "   |";
    $k = 0;
    for ($i=0; $i<6; $i++)
      { printf " %s", substr ($adfgvx,$i,1); }
    print "\n---+-------------\n";

    for ($i=0; $i<6; $i++)
      {
      printf " %s |", substr ($adfgvx,$i,1);
      for ($j=0; $j<6; $j++)
        {
        printf " %s", substr ($array,$k,1);
        $k++;
        }
      print "\n";
      }
    print "\n";
    }

# Common code for input/output. NOTE that valid plaintext and
#  ciphertext use a different set of valid characters.

  if ($action eq "Encrypt")
      { print "\nEnter message to encrypt (A-Z,0-9):\n"; }
    else
      { print "\nEnter message to decrypt (ADFGVX):\n"; }

  chomp ($msg_in=<>);

# Convert message to upper case and remove non-alpha/non-numeric characters

  $msg_in = uc ($msg_in);

  if ($action eq "Encrypt")
      {
      $msg_in =~ s/[^A-Z0-9]//g;

      $block = set_output_block_size ();

#     Take each character in $msg_in and convert to letter pairs in $msg_temp

      $msg_temp = "";
      for ($i=0; $i < length ($msg_in); $i++)
        {
        $j = -1;
        do
          {
          $j++;
          $k = substr ($array,$j,1);
          }
          until ($k eq substr ($msg_in,$i,1));

        $row = int ($j / 6);
        $col = $j % 6;
        $msg_temp = $msg_temp . substr ($adfgvx,$row,1) . substr ($adfgvx,$col,1);

        if ($verbose == 1)
          { printf "Char: %s Intermediate message: %s\n", substr ($msg_in,$i,1), lc ($msg_temp); }
        }

        if ($verbose == 1)
          {
          print "\n $keyword\n $mix\n";
          for ($i=0;$i<length($msg_temp)/$num_col;$i++)
            {
            print " ";
            for ($j=0;$j<$num_col;$j++)
              { printf "%s", substr ($msg_temp,$i*$num_col+$j,1); }
              print "\n";
            }
          }

      $msg_out = transpose ($msg_temp,$mix);

      print "\n\nEncrypted message = ";
      cipher_out ($msg_out,$block);
      print "\n";
      }
    else
      {
      $msg_in =~ s/[^ADFGVX]//g;

#     Recover intermediate message state from input message

      $msg_temp = untranspose ($msg_in,$mix);

      if ($verbose == 1)
        { print "\nIntermediate state = $msg_temp\n"; }

      $msg_out = "";
      for ($i = 0; $i < length($msg_in)/2; $i++)
        {
        $j = adfgvx_rowcol (substr ($msg_temp,$i*2,1));
        $k = adfgvx_rowcol (substr ($msg_temp,$i*2+1,1));
        $msg_out = $msg_out . substr ($array, $j*6+$k, 1);

        if ($verbose == 1)
          { printf "Cipher pair %s%s Row %d Col %d Plain text %s\n",
              substr ($msg_temp,$i*2,1), substr ($msg_temp,$i*2+1,1), $j, $k, $msg_out; }
        }

      print "\n\nDecrypted message = $msg_out\n";
      }
  }

# ********************************************************
# OneTimePad
# ********************************************************

# Implements a One-Time Pad encryption/decryption

# Very simple routine -- asks for pad and message to encrypt/decrypt.
# Add pad[i] to message[i] mod 26 to get output.

sub OneTimePad ($action)
  {
  my ($block, $i, $msg_in, $msglen, $msg_out, $newchr, $pad);
  my $action; $action = $_[0];

  print "\n ** One-Time Pad**\n\n";
  printf " == %sion ==\n", $action;

# Common code for setup and I/O

# Start by getting the transposition keyword and setting the column
#   transpostion sort order

  print "\nEnter one-time pad (A-Z): ";
  chomp ($pad = <STDIN>);
  $pad = uc ($pad);
  $pad =~ s/[^A-Z]//g;
  printf "\nEnter message to %s (A-Z): ", lc($action);
  chomp ($msg_in = <STDIN>);

# Set up message arrays and variables

  ($msg_in, $msg_out, $msglen) = set_msg_values ($msg_in);

  if (length($pad) < $msglen)
    {
    print "\nError -- Length of one-time pad shorter than message.\n";
    return;
    }

  if ($action eq "Encrypt")
      {
      $block = set_output_block_size ();
    
      for ($i=0; $i<$msglen; $i++)
        {
        $newchr = chr ((((ord (substr ($msg_in,$i,1))-$ordA+1) + (ord (substr ($pad,$i,1))-$ordA+1)) % 26) + $ordA - 1);

        # Handle special case where the new character should be "Z" (26) but mod 26 wraps it to "@" (0)
        if ($newchr eq chr ($ordA-1)) { $newchr = chr ($ordA + 25); }

        $msg_out = $msg_out . $newchr;
        if ($verbose == 1)
          { printf "[%d] plaintext %s (%d) + pad %s (%d) = ciphertext %s (%d)\n", $i, substr ($msg_in,$i,1), (ord (substr ($msg_in,$i,1))-$ordA+1),
              substr ($pad,$i,1), (ord (substr ($pad,$i,1))-$ordA+1), substr ($msg_out,$i,1), (ord (substr ($msg_out,$i,1))-$ordA+1); }
        }

      print "Encrypted message = ";
      cipher_out ($msg_out,$block);
      print "\n";
      }
    else
      {
      for ($i=0; $i<$msglen; $i++)
        {
        $newchr = chr ((((ord (substr ($msg_in,$i,1))-$ordA+1) - (ord (substr ($pad,$i,1))-$ordA+1)) % 26) + $ordA - 1);

        # Handle special case where the new character should be "Z" (26) but mod 26 wraps it to "@" (0)
        if ($newchr eq chr ($ordA-1)) { $newchr = chr ($ordA + 25); }

        $msg_out = $msg_out . $newchr;
        if ($verbose == 1)
          { printf "[%d] ciphertext %s (%d) - pad %s (%d) = plaintext %s (%d)\n", $i, substr ($msg_in,$i,1), (ord (substr ($msg_in,$i,1))-$ordA+1),
              substr ($pad,$i,1), (ord (substr ($pad,$i,1))-$ordA+1), substr ($msg_out,$i,1), (ord (substr ($msg_out,$i,1))-$ordA+1); }
        }

      print "Decrypted message = $msg_out\n";
      }
  }

# ********************************************************
# set_alphabet
# ********************************************************

# @alphabet is a 26-byte array with the letters in order of use in the Vigenere square.
# @alphanum is a 26-byte array indicating where the i-th letter is in the @alphabet
#   array; e.g., if @alphabet starts as ECABD..., then the first five entries of
#   @alphanum would be 2,3,1,4,0,...
# @AtoZ is a 26-byte array indicating whether the letters A-Z have been used (where
#   AtoZ [0] = A, AtoZ [1] = B, ..., AtoZ [25] = Z)

sub set_alphabet ($baseword)
  {
  my ($baselen, $ch, $i, $j, $n);
  my $baseword; $baseword = $_[0];

  for ($i=0; $i<=25; $i++)
    {
    $alphabet [$i] = "";
    $AtoZ [$i] = 0;
    }

# Process the Baseword if it is not null!

  $n = 0;
  $baselen = length ($baseword);
  if ($baselen > 0)
    {
    for ($i=0; $i<$baselen;$i++)
      {

# Get a character from the baseword...

      $ch = uc (substr ($baseword,$i,1));
      $j = ord ($ch) - $ordA;

# If the baseword letter has not yet been used, then add it to the alphabet and mark it used

      if (($j >= 0 && $j <= 25) && $AtoZ [$j] == 0)
        {
        $AtoZ [$j] = 1;
        $alphabet [$n] = $ch;
        $alphanum [$j] = $n;
        $n++;
        }
      }
    }

# At this point, fill the rest of the alphabet array with whatever letters have not yet been used

  if ($n < 25)
    {
    for ($i=$0; $i<26; $i++)
      {
      if ($AtoZ [$i] == 0)
        {
        $AtoZ [$i] = 1;
        $alphabet [$n] = chr ($ordA + $i);
        $alphanum [$i] = $n;
        $n++;
        }
      }
    }

# If verbose, print out the square

  if ($verbose == 1)
    {
    print "\nAlphabet with baseword = ";
    if ($baseword eq "")
        { print "NULL"; }
      else
        { print $baseword; }
    print "\n\n   |";
    for ($i=$0; $i<26; $i++)
      { print " $alphabet[$i]"; }
    print "\n";

    print "---+-----------------------------------------------------\n";

    for ($i=$0; $i<26; $i++)
      {
      print " $alphabet[$i] |";
      for ($j=$0; $j<26; $j++)
        {
        $n = ($i + $j) % 26;
        print " $alphabet[$n]";
        }
      print "\n";
      }
    }
  }


# ********************************************************
# adfgvx_rowcol
# ********************************************************

# Simple routine to find out what position $char is in the
# "ADFGVX" string

sub adfgvx_rowcol ($char)
  {
  my ($i, $test);
  my $char; $char = $_[0];

  $i = -1;
  do
    {
    $i++;
    $test = substr ("ADFGVX",$i,1);
    }
    until ($test eq $char);

  return ($i);
  }

# ********************************************************
# Transposition
# ********************************************************

# This routine handles columnar and Myszkowski transpositions.
# The difference between the two is that columnar just outputs
# the columns from left to right in sequence while
# Myszkowski uses a keyword to sort the column order

# For columnar transposition cipher, ask for the number of columns
#   $mix is then 0123...
# For Myszkowski, ask for the keyword and compute $mix

sub Transposition ($action, $type)
  {
  my ($i, $num_col, $mix, $msg_in, $msg_out, $block, $keyword);
  my ($action, $type); $action = $_[0]; $type = $_[1];

  print "\n ** $type Transposition Operation**\n\n";
  printf " == %sion ==\n\n", $action;

# Get number of columns/keyword for transposition

  if ($type eq "Columnar")
      {
      print "Enter number of columns: ";
      chomp ($num_col = <STDIN>);
      $mix = "";
      for ($i = 0; $i < $num_col; $i++)
        { $mix = $mix . chr ($i + $ord0); }
      }
    else

# Code for Myszkowski transposition

      {
      print "Enter transposition keyword: ";
      chomp ($keyword = <STDIN>);
      $keyword = uc ($keyword);
      $keyword =~ s/[^A-Z]//g;
      $num_col = length ($keyword);
      $mix = sort_key ($keyword);
      }

# Common code for reading plaintext/ciphertext

  printf "\nEnter message to %s\n", lc ($action);
  chomp ($msg_in=<>);

# Set up message arrays and variables

  ($msg_in, $msg_out, $msglen) = set_msg_values ($msg_in);

  if ($action eq "Encrypt")
      {
      $block = set_output_block_size ();

      $msg_out = transpose ($msg_in,$mix);

# Candidate verbose text

# print "message $msg_in order $mix numcols $num_col\n";
# print "matrix:\n";
# for ($xx=0;$xx<length($msg_in);$xx++)
#   {
#   printf "%s", substr($msg_in,$xx,1);
#   if ((($xx+1) % $num_col) == 0)
#     { print "\n"; }
#   }
# print "\n";

      print "\n\nEncrypted message = ";
      cipher_out ($msg_out,$block);
      print "\n";
      }
    else
      {
      $msg_out = untranspose ($msg_in,$mix);
      print "\n\nDecrypted message = $msg_out\n";
      }
  }

# ********************************************************
# transpose
# ********************************************************

# Input $msg_in with a sort order ($order) and number of columns ($width)

# Transpose the $width columns of $msg_in using the sequence in $order,
# to produce $msg_out

# If $n = 1, add random characters A-Z to fill array; if $n =2, choose
# random characters from sequence "ADFGVX"

sub transpose ($msg_in,$order)
  {
  my (@array, $i, $j, $k, $row, $col, $msg_out, $width);
  my ($msg_in, $order); $msg_in = $_[0]; $order = $_[1];

$n = 1;

# Put $msg_in into $array with $width columns

  $width = length ($order);
  $row = 0; $col = 0;
  for ($i = 0; $i < length ($msg_in); $i++)
    {
    $array [$row][$col] = substr ($msg_in,$i,1);
    $col++;
    if ($col % $width == 0)
      {
      $col = 0;
      $row++;
      }
    }

  if ($col == 0)
    { $row--; }

  if ($verbose == 1)
    {
    print "\nGrid...\n";
    for ($i=0; $i<=$row; $i++)
      {
      for ($j=0; $j<$width; $j++)
        { printf "%s", $array [$i][$j]; }
      print "\n";
      }
    print "\n";
    }

# Transpose $array to $msg_out

  $msg_out = "";
  for ($j = 0; $j < $width; $j++)
    {
    $k = ord (substr ($order,$j,1)) - $ord0;
    for ($i = 0; $i <= $row; $i++)
      { $msg_out = $msg_out . $array [$i][$k]; }
    }

  return ($msg_out);
  }

# ********************************************************
# untranspose
# ********************************************************

# Reverses the transposition of a message block using the
# column sequence in $order.

sub untranspose ($msg_in,$order)
  {
  my (@array, $i, $j, $k, $n, $row, $col, $msg_out, $width);
  my ($msg_in, $order, $num_bytes); $msg_in = $_[0]; $order = $_[1];

# The trick for this routine is that the number of characters might be an even
#   multiple of #_rows x width. If it were, cryptanalysis would be greatly
#   simplified; if you know the length of the input, you could make some
#   guesses as to number of columns. So, we need to account for the last
#   row of the letter grid not having $width number of characters

  $width = length ($order);

# $num_rows is number of complete rows

  $num_rows = int (length ($msg_in) / $width);

# $last_row is number of characters in the last row. If this value = 0, then
#   the grid is a full rectangle; if != 0, then there are "extra" characters
#   to accomodate in the first $last_row number of columns

  $last_row = length ($msg_in) % $width;

# Create $array from $msg_in, $order and $width (so, we recover and
#  unscamble all at once)

  $n = 0;
  for ($j = 0; $j < $width; $j++)
    {
    $k = ord (substr ($order,$j,1)) - $ord0;
    for ($i = 0; $i < $num_rows; $i++)
      {
      $array [$i][$k] = substr ($msg_in,$n,1);
      $n++;
      }
    if ($last_row != 0 && $k < $last_row)
      {
      $array [$num_rows+1][$k] = substr ($msg_in,$n,1);
      $n++;
      }
    }

  if ($verbose == 1)
    {
    print "\nGrid...\n";
    for ($i=0; $i<$num_rows; $i++)
      {
      for ($j=0; $j<$width; $j++)
        { printf "%s", $array [$i][$j]; }
      print "\n";
      }
    if ($last_row != 0)
      {
      for ($j=0; $j<$last_row; $j++)
        { printf "%s", $array [$num_rows+1][$j]; }
      print "\n";
      }
    print "\n";
    }

# Create $msg_out from $array

  $msg_out = "";
  for ($i = 0; $i < $num_rows; $i++)
    {
    for ($j = 0; $j < $width; $j++)
      {
      $msg_out = $msg_out . $array [$i][$j];
      }
    }
  if ($last_row != 0)
    {
    for ($j=0; $j<$last_row; $j++)
      { $msg_out = $msg_out . $array [$num_rows+1][$j]; }
    print "\n";
    }

  return ($msg_out);
  }

# ********************************************************
# cipher_out
# ********************************************************

# Accepts a string of $ciphertext and displays in in blocks
#   $block bytes in length.

# If $block=0, just display the ciphertext without spaces.

sub cipher_out ($ciphertext,$block)
  {
  my ($i, $j);
  my ($ciphertext, $block);
  $ciphertext = $_[0]; $block = $_[1];

# Take $ciphertext and convert to lowercase and print in $block-byte blocks.

  if ($block == 0)
    {
    printf "%s", lc($ciphertext);
    return;
    }

  $j = 0;

  for ($i = 0; $i < length ($ciphertext); $i++)
    {
    printf "%s", lc (substr ($ciphertext,$i,1));
    $j = ($j + 1) % $block;
    if ($j == 0)
      { print " "; }
    }
  }

# ********************************************************
# sort_key
# ********************************************************

# Input $key and produce $order, a scalar string with the
# alphabetical/numeric ordering of the characters comprising
# $key.

# $key is assumed to be upper-case alpha

# E.g.: $key = "KESSLER" would yield $order = "1504623"

sub sort_key ($key)
  {
  my (@array, $i, $j, $low, $lownum, $n, $order);
  my $key; $key = $_[0];

  $order = "";
  $n = length ($key);
  for ($i=0; $i<$n; $i++)
    { $array [$i] = substr ($key,$i,1) }

  for ($i=0; $i<$n; $i++)
    {
    $low = chr(96);
    for ($j=0; $j<$n; $j++)
      {
      if (ord($array [$j]) < ord($low))
        {
        $low = $array [$j];
        $lownum = $j;
        }
      }
      $order = $order . chr ($lownum + $ord0);
      $array [$lownum] = chr(96);
    }

  return ($order);
  }

# ********************************************************
# set_msg_values
# ********************************************************

# This routine replaces eight occurrences of common code in the
# program. Its four lines do four things:

# 1) Convert the input message to upper case
# 2) Remove non-alpha characters from the input message
# 3) Create a variabel to store the input message length
# 4) Initialize the output message array 

sub set_msg_values ($msg_in)
  {
  my ($msg_out, $msglen);
  my $msg_in; $msg_in = $_[0];

  $msg_in = uc ($msg_in);
  $msg_in =~ s/[^A-Z]//g;
  $msglen = length ($msg_in);
  $msg_out = "";

  return ($msg_in, $msg_out, $msglen);
  }

# ********************************************************
# set_output_block_size
# ********************************************************

# This routine replaces eight occurrences of common code in the
# program. Its sole function is to read in the output block size
# and ensure that the value is an integer value between 0 and 10

sub set_output_block_size ()
  {
  my $block;

  do
    {
    print "\nEnter block size for ciphertext (0-10, where 0 means single string; 3-5 recommended):\n";
    chomp ($block=<>);
    }
    until ($block =~ /^[0-9]*$/ && $block >= 0 && $block <=10);

    print "\n";

  return ($block);
  }

